"""Kedro Telemetry plugin for collecting Kedro usage data."""

from __future__ import annotations

import hashlib
import json
import logging
import os
import sys
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any

import requests
import tomli_w
import yaml
from appdirs import user_config_dir
from kedro import __version__ as KEDRO_VERSION
from kedro.framework.cli.cli import KedroCLI
from kedro.framework.cli.hooks import cli_hook_impl
from kedro.framework.hooks import hook_impl
from kedro.framework.project import PACKAGE_NAME, pipelines
from kedro.framework.startup import ProjectMetadata
from kedro.io.data_catalog import DataCatalog
from kedro.pipeline import Pipeline

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

from kedro_telemetry import __version__ as TELEMETRY_VERSION
from kedro_telemetry.masking import _mask_kedro_cli

HEAP_APPID_PROD = "2388822444"
HEAP_ENDPOINT = "https://heapanalytics.com/api/track"
HEAP_HEADERS = {"Content-Type": "application/json"}
KNOWN_CI_ENV_VAR_KEYS = {
    "GITLAB_CI",  # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
    "GITHUB_ACTION",  # https://docs.github.com/en/actions/learn-github-actions/variables#default-environment-variables
    "BITBUCKET_BUILD_NUMBER",  # https://support.atlassian.com/bitbucket-cloud/docs/variables-and-secrets/
    "JENKINS_URL",  # https://www.jenkins.io/doc/book/pipeline/jenkinsfile/#using-environment-variables
    "CODEBUILD_BUILD_ID",  # https://docs.aws.amazon.com/codebuild/latest/userguide/build-env-ref-env-vars.html
    "CIRCLECI",  # https://circleci.com/docs/variables/#built-in-environment-variables
    "TRAVIS",  # https://docs.travis-ci.com/user/environment-variables/#default-environment-variables
    "BUILDKITE",  # https://buildkite.com/docs/pipelines/environment-variables
}
_SKIP_TELEMETRY_ENV_VAR_KEYS = (
    "DO_NOT_TRACK",
    "KEDRO_DISABLE_TELEMETRY",
)
TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
CONFIG_FILENAME = "telemetry.toml"
PYPROJECT_CONFIG_NAME = "pyproject.toml"
UNDEFINED_PACKAGE_NAME = "undefined_package_name"
MISSING_USER_IDENTITY = "missing_user_identity"

logger = logging.getLogger(__name__)


def _hash(string: str) -> str:
    return hashlib.sha512(bytes(string, encoding="utf8")).hexdigest()


def _get_or_create_uuid() -> str:
    """
    Reads a UUID from a configuration file or generates and saves a new one if not present.
    """
    config_path = user_config_dir("kedro")
    full_path = os.path.join(config_path, CONFIG_FILENAME)

    try:
        if os.path.exists(full_path):
            with open(full_path, "rb") as f:
                config = tomllib.load(f)

                if "telemetry" in config and "uuid" in config["telemetry"]:
                    return uuid.UUID(config["telemetry"]["uuid"]).hex

        # Generate a new UUID and save it to the config file
        new_uuid = _generate_new_uuid(full_path)

        return new_uuid

    except Exception as e:
        logging.debug(f"Failed to retrieve UUID: {e}")
        return ""


def _get_or_create_project_id(pyproject_path: Path) -> str | None:
    """
    Reads a project id from a configuration file or generates and saves a new one if not present.
    Returns None if configuration file does not exist or does not relate to Kedro.
    """
    try:
        with open(pyproject_path, "r+b") as file:
            pyproject_data = tomllib.load(file)

            # Check if pyproject related to kedro
            try:
                _ = pyproject_data["tool"]["kedro"]
                try:
                    project_id = pyproject_data["tool"]["kedro_telemetry"]["project_id"]
                except KeyError:
                    project_id = uuid.uuid4().hex
                    toml_string = (
                        f'\n[tool.kedro_telemetry]\nproject_id = "{project_id}"\n'
                    )
                    file.write(toml_string.encode("utf-8"))
                return project_id
            except KeyError:
                logging.debug(
                    f"Failed to retrieve project id or save project id: "
                    f"{str(pyproject_path)} does not contain a [tool.kedro] section"
                )
                return None
    except OSError as exc:
        logging.debug(f"Failed to read the file: {str(pyproject_path)}.\n{str(exc)}")
    return None


def _add_tool_properties(
    properties: dict[str, Any], pyproject_path: Path
) -> dict[str, Any]:
    """
    Extends project properties with tool's properties.
    """
    if pyproject_path.exists():
        with open(pyproject_path, "rb") as file:
            pyproject_data = tomllib.load(file)

        try:
            tool_kedro = pyproject_data["tool"]["kedro"]
            if "tools" in tool_kedro:
                properties["tools"] = ", ".join(tool_kedro["tools"])
            if "example_pipeline" in tool_kedro:
                properties["example_pipeline"] = tool_kedro["example_pipeline"]
        except KeyError:
            pass

    return properties


def _generate_new_uuid(full_path: str) -> str:
    try:
        config: dict[str, dict[str, Any]] = {"telemetry": {}}
        new_uuid = uuid.uuid4().hex
        config["telemetry"]["uuid"] = new_uuid

        os.makedirs(os.path.dirname(full_path), exist_ok=True)
        with open(full_path, "wb") as f:
            tomli_w.dump(config, f)

        return new_uuid
    except Exception as e:
        logging.debug(f"Failed to create UUID: {e}")
        return ""


class KedroTelemetryHook:
    """Hook to send CLI command data to Heap"""

    def __init__(self):
        self._consent = None
        self._sent = False
        self._event_properties = None
        self._project_path = None
        self._user_uuid = None

    @cli_hook_impl
    def before_command_run(
        self, project_metadata: ProjectMetadata, command_args: list[str]
    ):
        """Hook implementation to send command run data to Heap"""

        project_path = project_metadata.project_path if project_metadata else None

        consent = _check_for_telemetry_consent(project_path)
        self._consent_is_implicit = consent is None
        self._consent = self._consent_is_implicit or consent
        if not self._consent:
            return

        # get KedroCLI and its structure from actual project root
        cli = KedroCLI(project_path=project_path if project_path else Path.cwd())
        masked_command_args = _mask_kedro_cli(cli, command_args=command_args)

        self._user_uuid = _get_or_create_uuid()

        event_properties = _get_project_properties(self._user_uuid, project_path)
        event_properties["command"] = (
            f"kedro {' '.join(masked_command_args)}" if masked_command_args else "kedro"
        )
        event_properties["main_command"] = (
            masked_command_args[0] if masked_command_args else "kedro"
        )

        self._event_properties = event_properties

    @cli_hook_impl
    def after_command_run(self):
        if self._consent and not self._sent:
            self._send_telemetry_heap_event("CLI command")

    @hook_impl
    def after_context_created(self, context):
        """Hook implementation to read metadata"""

        consent = _check_for_telemetry_consent(context.project_path)
        self._consent_is_implicit = consent is None
        self._consent = self._consent_is_implicit or consent
        self._project_path = context.project_path

    @hook_impl
    def after_catalog_created(self, catalog):
        """Hook implementation to send project statistics data to Heap"""

        if self._consent is False:
            return

        default_pipeline = pipelines.get("__default__")  # __default__

        if not self._user_uuid:
            self._user_uuid = _get_or_create_uuid()

        if not self._event_properties:
            self._event_properties = _get_project_properties(
                self._user_uuid, self._project_path
            )

        project_properties = _format_project_statistics_data(
            catalog, default_pipeline, pipelines
        )
        self._event_properties.update(project_properties)

        self._send_telemetry_heap_event("Kedro Project Statistics")

    def _send_telemetry_heap_event(self, event_name: str):
        """Hook implementation to send command run data to Heap"""

        if self._consent_is_implicit:
            logger.info(
                "Kedro is sending anonymous usage data with the sole purpose of improving the product. "
                "No personal data or IP addresses are stored on our side. "
                "To opt out, set the `KEDRO_DISABLE_TELEMETRY` or `DO_NOT_TRACK` environment variables, "
                "or create a `.telemetry` file in the current working directory with the contents `consent: false`. "
                "To hide this message, explicitly grant or deny consent. "
                "Read more at https://docs.kedro.org/en/stable/configuration/telemetry.html"
            )

        try:
            _send_heap_event(
                event_name=event_name,
                identity=self._user_uuid if self._user_uuid else MISSING_USER_IDENTITY,
                properties=self._event_properties,
            )
            self._sent = True
        except Exception as exc:
            logger.debug(
                "Something went wrong in hook implementation to send command run data to Heap. "
                "Exception: %s",
                exc,
            )


def _is_known_ci_env(known_ci_env_var_keys: set[str]):
    # Most CI tools will set the CI environment variable to true
    if os.getenv("CI") == "true":
        return True
    # Not all CI tools follow this convention, we can check through those that don't
    return any(os.getenv(key) for key in known_ci_env_var_keys)


def _get_project_properties(user_uuid: str, project_path: Path | None) -> dict:
    if project_path:
        pyproject_path = project_path / PYPROJECT_CONFIG_NAME
        project_id = _get_or_create_project_id(pyproject_path)
        package_name = PACKAGE_NAME or UNDEFINED_PACKAGE_NAME
        hashed_project_id = (
            _hash(f"{project_id}{package_name}") if project_id is not None else None
        )
    else:
        hashed_project_id = None

    properties = {
        "username": user_uuid,
        "project_id": hashed_project_id,
        "project_version": KEDRO_VERSION,
        "telemetry_version": TELEMETRY_VERSION,
        "python_version": sys.version,
        "os": sys.platform,
        "is_ci_env": _is_known_ci_env(KNOWN_CI_ENV_VAR_KEYS),
    }

    if project_path:
        properties = _add_tool_properties(properties, pyproject_path)

    return properties


def _format_project_statistics_data(
    catalog: DataCatalog,
    default_pipeline: Pipeline,
    project_pipelines: dict,
):
    """Add project statistics to send to Heap."""
    # Support both catalog.list() for `kedro < 1.0` and catalog.keys() for `kedro >= 1.0`
    if hasattr(catalog, "keys") and callable(catalog.keys):
        dataset_names = catalog.keys()
    else:
        dataset_names = catalog.list()  # type: ignore

    project_statistics_properties = {}
    project_statistics_properties["number_of_datasets"] = sum(
        1
        for c in dataset_names
        if not c.startswith("parameters") and not c.startswith("params:")
    )
    project_statistics_properties["number_of_nodes"] = (
        len(default_pipeline.nodes) if default_pipeline else None  # type: ignore
    )
    project_statistics_properties["number_of_pipelines"] = len(project_pipelines.keys())
    return project_statistics_properties


def _get_heap_app_id() -> str:
    """
    Get the Heap App ID to send the data to.
    This will be the development ID if it's set as an
    environment variable, otherwise it will be the production ID.
    """
    return os.environ.get("HEAP_APPID_DEV", HEAP_APPID_PROD)


def _send_heap_event(
    event_name: str, identity: str, properties: dict[str, Any] | None = None
) -> None:
    data = {
        "app_id": _get_heap_app_id(),
        "event": event_name,
        "timestamp": datetime.now().strftime(TIMESTAMP_FORMAT),
        "properties": properties or {},
        "identity": identity,
    }

    try:
        resp = requests.post(
            url=HEAP_ENDPOINT, headers=HEAP_HEADERS, data=json.dumps(data), timeout=10
        )
        if resp.status_code != 200:  # noqa: PLR2004
            logger.debug(
                "Failed to send data to Heap. Response code returned: %s, Response reason: %s",
                resp.status_code,
                resp.reason,
            )
    except requests.exceptions.RequestException as exc:
        logger.debug(
            "Failed to send data to Heap. Exception of type '%s' was raised.",
            type(exc).__name__,
        )


def _check_for_telemetry_consent(project_path: Path | None) -> bool | None:
    """
    Use telemetry consent from ".telemetry" file if it exists and has a valid format.
    Telemetry is considered as opt-in otherwise.
    """

    for env_var in _SKIP_TELEMETRY_ENV_VAR_KEYS:
        if os.environ.get(env_var):
            return False

    if project_path:
        telemetry_file_path = project_path / ".telemetry"
        if telemetry_file_path.exists():
            with open(telemetry_file_path, encoding="utf-8") as telemetry_file:
                telemetry = yaml.safe_load(telemetry_file)
                if _is_valid_syntax(telemetry):
                    return telemetry["consent"]
    return None


def _is_valid_syntax(telemetry: Any) -> bool:
    return isinstance(telemetry, dict) and isinstance(
        telemetry.get("consent", None), bool
    )


telemetry_hook = KedroTelemetryHook()



import logging
import sys
from pathlib import Path
from unittest.mock import MagicMock

import requests
import yaml
from kedro import __version__ as kedro_version
from kedro.framework.project import pipelines
from kedro.framework.startup import ProjectMetadata
from kedro.io import DataCatalog, MemoryDataset
from kedro.pipeline import Pipeline, node
from kedro.pipeline import pipeline as modular_pipeline
from pytest import fixture, mark

from kedro_telemetry import __version__ as TELEMETRY_VERSION
from kedro_telemetry.plugin import (
    _SKIP_TELEMETRY_ENV_VAR_KEYS,
    KNOWN_CI_ENV_VAR_KEYS,
    MISSING_USER_IDENTITY,
    KedroTelemetryHook,
    _check_for_telemetry_consent,
    _format_project_statistics_data,
    _is_known_ci_env,
)

REPO_NAME = "dummy_project"
PACKAGE_NAME = "dummy_package"

MOCK_PYPROJECT_TOOLS = """
[build-system]
requires = [ "setuptools",]
build-backend = "setuptools.build_meta"

[project]
name = "spaceflights"
readme = "README.md"
dynamic = [ "dependencies", "version",]

[project.scripts]
new-proj = "spaceflights.__main__:main"

[tool.kedro]
package_name = "spaceflights"
project_name = "spaceflights"
kedro_init_version = "0.18.14"
tools = ["Linting", "Testing", "Custom Logging", "Documentation", "Data Structure", "PySpark"]
example_pipeline = "True"

[project.entry-points."kedro.hooks"]

[tool.setuptools.dynamic.dependencies]
file = "requirements.txt"

[tool.setuptools.dynamic.version]
attr = "spaceflights.__version__"

[tool.setuptools.packages.find]
where = [ "src",]
namespaces = false
"""


@fixture
def fake_metadata(tmp_path):
    metadata = ProjectMetadata(
        config_file=tmp_path / REPO_NAME / "pyproject.toml",
        package_name=PACKAGE_NAME,
        project_name="CLI Testing Project",
        project_path=tmp_path / REPO_NAME,
        source_dir=tmp_path / REPO_NAME / "src",
        kedro_init_version=kedro_version,
        tools=[],
        example_pipeline="No",
    )
    return metadata


@fixture
def fake_catalog():
    catalog = DataCatalog(
        {
            "dummy_1": MemoryDataset(),
            "dummy_2": MemoryDataset(),
            "dummy_3": MemoryDataset(),
            "parameters": MemoryDataset(),
            "params:dummy": MemoryDataset(),
        }
    )
    return catalog


def identity(arg):
    return arg


@fixture
def fake_context():
    class MockKedroContext:
        # A dummy stand-in for KedroContext sufficient for this test
        project_path = Path("")

    return MockKedroContext()


@fixture
def fake_default_pipeline():
    mock_default_pipeline = modular_pipeline(
        [
            node(identity, ["input"], ["intermediate"], name="node0"),
            node(identity, ["intermediate"], ["output"], name="node1"),
        ],
    )
    return mock_default_pipeline


@fixture
def fake_sub_pipeline():
    mock_sub_pipeline = modular_pipeline(
        [
            node(identity, ["input"], ["intermediate"], name="node0"),
        ],
    )
    return mock_sub_pipeline


@fixture
def pipeline_fixture() -> Pipeline:
    mock_pipeline = MagicMock(spec=Pipeline)
    mock_pipeline.nodes = ["node1", "node2"]
    return mock_pipeline


@fixture
def project_pipelines() -> dict[str, Pipeline]:
    return {
        "pipeline1": MagicMock(spec=Pipeline),
        "pipeline2": MagicMock(spec=Pipeline),
    }


class TestKedroTelemetryHook:
    def test_before_command_run(self, mocker, fake_metadata, caplog):
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=None
        )
        mocker.patch("kedro_telemetry.plugin._is_known_ci_env", return_value=True)
        mocked_anon_id = mocker.patch("kedro_telemetry.plugin._hash")
        mocked_anon_id.return_value = "digested"
        mocker.patch("kedro_telemetry.plugin.PACKAGE_NAME", "spaceflights")
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_uuid",
            return_value="user_uuid",
        )
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_project_id",
            return_value="project_id",
        )

        mocked_heap_call = mocker.patch("kedro_telemetry.plugin._send_heap_event")

        with caplog.at_level(logging.INFO):
            telemetry_hook = KedroTelemetryHook()
            command_args = ["--version"]
            telemetry_hook.before_command_run(fake_metadata, command_args)
            telemetry_hook.after_command_run()
        expected_properties = {
            "username": "user_uuid",
            "project_id": "digested",
            "project_version": kedro_version,
            "telemetry_version": TELEMETRY_VERSION,
            "python_version": sys.version,
            "os": sys.platform,
            "is_ci_env": True,
            "command": "kedro --version",
        }
        generic_properties = {
            **expected_properties,
            "main_command": "--version",
        }

        expected_calls = [
            mocker.call(
                event_name="CLI command",
                identity="user_uuid",
                properties=generic_properties,
            ),
        ]
        assert mocked_heap_call.call_args_list == expected_calls
        assert any(
            "Kedro is sending anonymous usage data with the sole purpose of improving the product. "
            "No personal data or IP addresses are stored on our side. "
            "To opt out, set the `KEDRO_DISABLE_TELEMETRY` or `DO_NOT_TRACK` environment variables, "
            "or create a `.telemetry` file in the current working directory with the contents `consent: false`. "
            "To hide this message, explicitly grant or deny consent. "
            "Read more at https://docs.kedro.org/en/stable/configuration/telemetry.html"
            in record.message
            for record in caplog.records
        )

    def test_before_command_run_with_tools(self, mocker, fake_metadata):
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=True
        )
        mocker.patch("kedro_telemetry.plugin._is_known_ci_env", return_value=True)
        mocked_anon_id = mocker.patch("kedro_telemetry.plugin._hash")
        mocked_anon_id.return_value = "digested"
        mocker.patch("kedro_telemetry.plugin.PACKAGE_NAME", "spaceflights")
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_uuid",
            return_value="user_uuid",
        )
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_project_id",
            return_value="project_id",
        )

        mocked_heap_call = mocker.patch("kedro_telemetry.plugin._send_heap_event")
        mocker.patch(
            "builtins.open",
            mocker.mock_open(read_data=MOCK_PYPROJECT_TOOLS.encode("utf-8")),
        )
        mocker.patch("pathlib.Path.exists", return_value=True)
        telemetry_hook = KedroTelemetryHook()
        command_args = ["--version"]
        telemetry_hook.before_command_run(fake_metadata, command_args)
        telemetry_hook.after_command_run()
        expected_properties = {
            "username": "user_uuid",
            "project_id": "digested",
            "project_version": kedro_version,
            "telemetry_version": TELEMETRY_VERSION,
            "python_version": sys.version,
            "os": sys.platform,
            "is_ci_env": True,
            "command": "kedro --version",
            "tools": "Linting, Testing, Custom Logging, Documentation, Data Structure, PySpark",
            "example_pipeline": "True",
        }
        generic_properties = {
            **expected_properties,
            "main_command": "--version",
        }

        expected_calls = [
            mocker.call(
                event_name="CLI command",
                identity="user_uuid",
                properties=generic_properties,
            ),
        ]
        assert mocked_heap_call.call_args_list == expected_calls

    def test_before_command_run_empty_args(self, mocker, fake_metadata):
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=True
        )
        mocker.patch("kedro_telemetry.plugin._is_known_ci_env", return_value=True)
        mocked_anon_id = mocker.patch("kedro_telemetry.plugin._hash")
        mocked_anon_id.return_value = "digested"
        mocker.patch("kedro_telemetry.plugin.PACKAGE_NAME", "spaceflights")
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_uuid",
            return_value="user_uuid",
        )
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_project_id",
            return_value="project_id",
        )

        mocked_heap_call = mocker.patch("kedro_telemetry.plugin._send_heap_event")
        telemetry_hook = KedroTelemetryHook()
        command_args = []
        telemetry_hook.before_command_run(fake_metadata, command_args)
        telemetry_hook.after_command_run()
        expected_properties = {
            "username": "user_uuid",
            "project_id": "digested",
            "project_version": kedro_version,
            "telemetry_version": TELEMETRY_VERSION,
            "python_version": sys.version,
            "os": sys.platform,
            "is_ci_env": True,
            "command": "kedro",
        }
        generic_properties = {
            "main_command": "kedro",
            **expected_properties,
        }

        expected_calls = [
            mocker.call(
                event_name="CLI command",
                identity="user_uuid",
                properties=generic_properties,
            ),
        ]

        assert mocked_heap_call.call_args_list == expected_calls

    def test_before_command_run_no_consent_given(self, mocker, fake_metadata, caplog):
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=False
        )

        mocked_heap_call = mocker.patch("kedro_telemetry.plugin._send_heap_event")
        with caplog.at_level(logging.INFO):
            telemetry_hook = KedroTelemetryHook()
            command_args = ["--version"]
            telemetry_hook.before_command_run(fake_metadata, command_args)

        mocked_heap_call.assert_not_called()
        assert not any(
            "Kedro is sending anonymous usage data with the sole purpose of improving the product. "
            "No personal data or IP addresses are stored on our side. "
            "To opt out, set the `KEDRO_DISABLE_TELEMETRY` or `DO_NOT_TRACK` environment variables, "
            "or create a `.telemetry` file in the current working directory with the contents `consent: false`. "
            "To hide this message, explicitly grant or deny consent. "
            "Read more at https://docs.kedro.org/en/stable/configuration/telemetry.html"
            in record.message
            for record in caplog.records
        )

    def test_before_command_run_connection_error(self, mocker, fake_metadata, caplog):
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=True
        )
        telemetry_hook = KedroTelemetryHook()
        command_args = ["--version"]

        mocked_post_request = mocker.patch(
            "requests.post", side_effect=requests.exceptions.ConnectionError()
        )
        telemetry_hook.before_command_run(fake_metadata, command_args)
        telemetry_hook.after_command_run()
        msg = "Failed to send data to Heap. Exception of type 'ConnectionError' was raised."
        assert msg in caplog.messages[-1]
        mocked_post_request.assert_called()

    def test_before_command_run_anonymous(self, mocker, fake_metadata):
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=True
        )
        mocker.patch("kedro_telemetry.plugin._is_known_ci_env", return_value=True)
        mocked_anon_id = mocker.patch("kedro_telemetry.plugin._hash")
        mocked_anon_id.return_value = "digested"
        mocker.patch("kedro_telemetry.plugin.PACKAGE_NAME", "spaceflights")
        mocker.patch("builtins.open", side_effect=OSError)

        mocked_heap_call = mocker.patch("kedro_telemetry.plugin._send_heap_event")
        telemetry_hook = KedroTelemetryHook()
        command_args = ["--version"]
        telemetry_hook.before_command_run(fake_metadata, command_args)
        telemetry_hook.after_command_run()
        expected_properties = {
            "username": "",
            "command": "kedro --version",
            "project_id": None,
            "project_version": kedro_version,
            "telemetry_version": TELEMETRY_VERSION,
            "python_version": sys.version,
            "os": sys.platform,
            "is_ci_env": True,
        }
        generic_properties = {
            "main_command": "--version",
            **expected_properties,
        }

        expected_calls = [
            mocker.call(
                event_name="CLI command",
                identity=MISSING_USER_IDENTITY,
                properties=generic_properties,
            ),
        ]
        assert mocked_heap_call.call_args_list == expected_calls

    def test_before_command_run_heap_call_error(self, mocker, fake_metadata, caplog):
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=True
        )
        mocked_heap_call = mocker.patch(
            "kedro_telemetry.plugin._send_heap_event", side_effect=Exception
        )
        telemetry_hook = KedroTelemetryHook()
        command_args = ["--version"]

        telemetry_hook.before_command_run(fake_metadata, command_args)
        telemetry_hook.after_command_run()
        msg = (
            "Something went wrong in hook implementation to send command run data to"
            " Heap. Exception:"
        )
        assert msg in caplog.messages[-1]
        mocked_heap_call.assert_called()

    def test_check_for_telemetry_consent_given(self, mocker, fake_metadata):
        Path(fake_metadata.project_path, "conf").mkdir(parents=True)
        telemetry_file_path = fake_metadata.project_path / ".telemetry"
        with open(telemetry_file_path, "w", encoding="utf-8") as telemetry_file:
            yaml.dump({"consent": True}, telemetry_file)

        assert _check_for_telemetry_consent(fake_metadata.project_path)

    def test_check_for_telemetry_consent_not_given(self, mocker, fake_metadata):
        Path(fake_metadata.project_path, "conf").mkdir(parents=True)
        telemetry_file_path = fake_metadata.project_path / ".telemetry"
        with open(telemetry_file_path, "w", encoding="utf-8") as telemetry_file:
            yaml.dump({"consent": False}, telemetry_file)

        assert not _check_for_telemetry_consent(fake_metadata.project_path)

    @mark.parametrize("env_var", _SKIP_TELEMETRY_ENV_VAR_KEYS)
    def test_check_for_telemetry_consent_skip_telemetry_with_env_var(
        self, monkeypatch, fake_metadata, env_var
    ):
        monkeypatch.setenv(env_var, "True")
        Path(fake_metadata.project_path, "conf").mkdir(parents=True)
        telemetry_file_path = fake_metadata.project_path / ".telemetry"
        with open(telemetry_file_path, "w", encoding="utf-8") as telemetry_file:
            yaml.dump({"consent": True}, telemetry_file)

        assert not _check_for_telemetry_consent(fake_metadata.project_path)

    def test_check_for_telemetry_consent_empty_file(self, mocker, fake_metadata):
        Path(fake_metadata.project_path, "conf").mkdir(parents=True)
        telemetry_file_path = fake_metadata.project_path / ".telemetry"

        with open(telemetry_file_path, "w", encoding="utf-8") as telemetry_file:
            yaml.dump({}, telemetry_file)

        assert _check_for_telemetry_consent(fake_metadata.project_path) is None

    def test_check_for_telemetry_consent_file_no_consent_field(
        self, mocker, fake_metadata
    ):
        Path(fake_metadata.project_path, "conf").mkdir(parents=True)
        telemetry_file_path = fake_metadata.project_path / ".telemetry"
        with open(telemetry_file_path, "w", encoding="utf8") as telemetry_file:
            yaml.dump({"nonsense": "bla"}, telemetry_file)

        assert _check_for_telemetry_consent(fake_metadata.project_path) is None

    def test_check_for_telemetry_consent_file_invalid_yaml(self, mocker, fake_metadata):
        Path(fake_metadata.project_path, "conf").mkdir(parents=True)
        telemetry_file_path = fake_metadata.project_path / ".telemetry"
        telemetry_file_path.write_text("invalid_ yaml")

        assert _check_for_telemetry_consent(fake_metadata.project_path) is None

    @mark.parametrize(
        "env_vars,result",
        [
            ({"CI": "true"}, True),
            ({"CI": "false"}, False),
            ({"CI": "false", "CODEBUILD_BUILD_ID": "Testing known CI env var"}, True),
            ({"JENKINS_URL": "Testing known CI env var"}, True),
            ({"CI": "false", "TRAVIS": "Testing known CI env var"}, True),
            ({"GITLAB_CI": "Testing known CI env var"}, True),
            ({"CI": "false", "CIRCLECI": "Testing known CI env var"}, True),
            (
                {"CI": "false", "BITBUCKET_BUILD_NUMBER": "Testing known CI env var"},
                True,
            ),
        ],
    )
    def test_check_is_known_ci_env(self, monkeypatch, env_vars, result):
        for env_var, env_var_value in env_vars.items():
            monkeypatch.setenv(env_var, env_var_value)

        known_ci_vars = KNOWN_CI_ENV_VAR_KEYS
        # Because our CI runs on Github Actions, this would always return True otherwise
        known_ci_vars.discard("GITHUB_ACTION")
        assert _is_known_ci_env(known_ci_vars) == result

    def test_after_context_created_without_kedro_run(  # noqa: PLR0913
        self,
        mocker,
        fake_catalog,
        fake_default_pipeline,
        fake_sub_pipeline,
        fake_context,
    ):
        mocker.patch.dict(
            pipelines, {"__default__": fake_default_pipeline, "sub": fake_sub_pipeline}
        )
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=True
        )
        mocker.patch("kedro_telemetry.plugin._is_known_ci_env", return_value=True)
        mocker.patch("kedro_telemetry.plugin._hash", return_value="digested")
        mocker.patch("kedro_telemetry.plugin.PACKAGE_NAME", "spaceflights")
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_uuid",
            return_value="user_uuid",
        )
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_project_id",
            return_value="project_id",
        )

        mocked_heap_call = mocker.patch("kedro_telemetry.plugin._send_heap_event")
        mocker.patch("kedro_telemetry.plugin.open")
        mocker.patch("kedro_telemetry.plugin.tomllib.load")
        mocker.patch("kedro_telemetry.plugin.tomli_w.dump")

        # Without CLI invoked - i.e. `session.run` in Jupyter/IPython
        telemetry_hook = KedroTelemetryHook()
        telemetry_hook.after_context_created(fake_context)
        telemetry_hook.after_catalog_created(fake_catalog)

        project_properties = {
            "username": "user_uuid",
            "project_id": "digested",
            "project_version": kedro_version,
            "telemetry_version": TELEMETRY_VERSION,
            "python_version": sys.version,
            "os": sys.platform,
            "is_ci_env": True,
        }
        project_statistics = {
            "number_of_datasets": 3,
            "number_of_nodes": 2,
            "number_of_pipelines": 2,
        }
        expected_properties = {**project_properties, **project_statistics}
        expected_call = mocker.call(
            event_name="Kedro Project Statistics",
            identity="user_uuid",
            properties=expected_properties,
        )

        # The 1st call is the Project Hook without CLI
        assert mocked_heap_call.call_args_list[0] == expected_call

    def test_after_context_created_with_kedro_run(  # noqa: PLR0913
        self,
        mocker,
        fake_catalog,
        fake_metadata,
        fake_default_pipeline,
        fake_sub_pipeline,
        fake_context,
    ):
        mocker.patch.dict(
            pipelines, {"__default__": fake_default_pipeline, "sub": fake_sub_pipeline}
        )
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=True
        )
        mocker.patch("kedro_telemetry.plugin._is_known_ci_env", return_value=True)
        mocker.patch("kedro_telemetry.plugin._hash", return_value="digested")
        mocker.patch("kedro_telemetry.plugin.PACKAGE_NAME", "spaceflights")
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_uuid",
            return_value="user_uuid",
        )
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_project_id",
            return_value="project_id",
        )
        mocked_heap_call = mocker.patch("kedro_telemetry.plugin._send_heap_event")
        mocker.patch("kedro_telemetry.plugin.tomllib.load")
        mocker.patch("kedro_telemetry.plugin.tomli_w.dump")
        # CLI run first
        telemetry_cli_hook = KedroTelemetryHook()
        command_args = ["--version"]
        telemetry_cli_hook.before_command_run(fake_metadata, command_args)

        # Follow by project run
        telemetry_hook = KedroTelemetryHook()
        telemetry_hook.after_context_created(fake_context)
        telemetry_hook.after_catalog_created(fake_catalog)

        project_properties = {
            "username": "user_uuid",
            "project_id": "digested",
            "project_version": kedro_version,
            "telemetry_version": TELEMETRY_VERSION,
            "python_version": sys.version,
            "os": sys.platform,
            "is_ci_env": True,
        }
        project_statistics = {
            "number_of_datasets": 3,
            "number_of_nodes": 2,
            "number_of_pipelines": 2,
        }
        expected_properties = {**project_properties, **project_statistics}

        expected_call = mocker.call(
            event_name="Kedro Project Statistics",
            identity="user_uuid",
            properties=expected_properties,
        )

        assert mocked_heap_call.call_args_list[0] == expected_call

    def test_after_context_created_with_kedro_run_and_tools(  # noqa: PLR0913
        self,
        mocker,
        fake_catalog,
        fake_metadata,
        fake_default_pipeline,
        fake_sub_pipeline,
        fake_context,
    ):
        mocker.patch.dict(
            pipelines, {"__default__": fake_default_pipeline, "sub": fake_sub_pipeline}
        )
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=True
        )
        mocker.patch("kedro_telemetry.plugin._is_known_ci_env", return_value=True)
        mocker.patch("kedro_telemetry.plugin._hash", return_value="digested")
        mocker.patch("kedro_telemetry.plugin.PACKAGE_NAME", "spaceflights")
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_uuid",
            return_value="user_uuid",
        )
        mocker.patch(
            "kedro_telemetry.plugin._get_or_create_project_id",
            return_value="project_id",
        )
        mocked_heap_call = mocker.patch("kedro_telemetry.plugin._send_heap_event")
        mocker.patch(
            "builtins.open",
            mocker.mock_open(read_data=MOCK_PYPROJECT_TOOLS.encode("utf-8")),
        )
        mocker.patch("pathlib.Path.exists", return_value=True)

        # CLI run first
        telemetry_cli_hook = KedroTelemetryHook()
        command_args = ["--version"]
        telemetry_cli_hook.before_command_run(fake_metadata, command_args)

        # Follow by project run
        telemetry_hook = KedroTelemetryHook()
        telemetry_hook.after_context_created(fake_context)
        telemetry_hook.after_catalog_created(fake_catalog)

        project_properties = {
            "username": "user_uuid",
            "project_id": "digested",
            "project_version": kedro_version,
            "telemetry_version": TELEMETRY_VERSION,
            "python_version": sys.version,
            "os": sys.platform,
            "is_ci_env": True,
            "tools": "Linting, Testing, Custom Logging, Documentation, Data Structure, PySpark",
            "example_pipeline": "True",
        }
        project_statistics = {
            "number_of_datasets": 3,
            "number_of_nodes": 2,
            "number_of_pipelines": 2,
        }
        expected_properties = {**project_properties, **project_statistics}

        expected_call = mocker.call(
            event_name="Kedro Project Statistics",
            identity="user_uuid",
            properties=expected_properties,
        )

        assert mocked_heap_call.call_args_list[0] == expected_call

    def test_after_context_created_no_consent_given(self, mocker):
        fake_context = mocker.Mock()
        mocker.patch(
            "kedro_telemetry.plugin._check_for_telemetry_consent", return_value=False
        )

        mocked_heap_call = mocker.patch("kedro_telemetry.plugin._send_heap_event")
        telemetry_hook = KedroTelemetryHook()
        telemetry_hook.after_context_created(fake_context)

        mocked_heap_call.assert_not_called()

    def test_old_catalog_with_list_method(self, pipeline_fixture, project_pipelines):
        # catalog.list() was replaces with catalog.keys() in `kedro >= 1.0`
        catalog = MagicMock()
        catalog.list.return_value = [
            "dataset1",
            "params:my_param",
            "dataset2",
            "parameters",
        ]

        # Ensure .keys is not present
        if hasattr(catalog, "keys"):
            del catalog.keys

        result = _format_project_statistics_data(
            catalog, pipeline_fixture, project_pipelines
        )

        ds_nodes_pipes_cnt = 2

        assert result["number_of_datasets"] == ds_nodes_pipes_cnt
        assert result["number_of_nodes"] == ds_nodes_pipes_cnt
        assert result["number_of_pipelines"] == ds_nodes_pipes_cnt

    def test_new_catalog_with_keys_method(self, pipeline_fixture, project_pipelines):
        # catalog.list() was replaces with catalog.keys() in `kedro >= 1.0`
        catalog = MagicMock()
        catalog.keys.return_value = [
            "datasetA",
            "params:global",
            "datasetB",
            "parameters",
        ]
        # Ensure .list is not present
        if hasattr(catalog, "list"):
            del catalog.list

        result = _format_project_statistics_data(
            catalog, pipeline_fixture, project_pipelines
        )

        ds_nodes_pipes_cnt = 2

        assert result["number_of_datasets"] == ds_nodes_pipes_cnt
        assert result["number_of_nodes"] == ds_nodes_pipes_cnt
        assert result["number_of_pipelines"] == ds_nodes_pipes_cnt



"""Module containing command masking functionality."""

from __future__ import annotations

from typing import Any

import click

MASK = "*****"


def _recurse_cli(
    cli_element: click.Command | (click.Group | (click.CommandCollection | None)),
    ctx: click.Context,
    io_dict: dict[str | None, Any],
    get_help: bool = False,
) -> None:
    """
    Code copied over from kedro.tools.cli to maintain backwards compatibility
    with previous versions of kedro (<0.17.5).

    Recursive function that checks the type of the command (key) and decides:
    1. In case of `click.Group` or `click.CommandCollection` (aggregate commands),
        the function collects the name and recurses one layer deeper
        for each sub-command.
    2. In case of `click.Command`, the terminus command has been reached. The function
        collects the name, parameters and args, flattens them and saves them as
        dictionary keys.
    Args:
        cli_element: CLI Collection as input for recursion, typically `KedroCLI`.
        ctx: Click Context, created by the wrapper function.
        io_dict: Input-output dictionary, mutated during the recursion.
        get_help: Boolean fork - allows either:
            raw structure - nested dictionary until final value of `None`
            help structure - nested dictionary where leaves are `--help` cmd output

    Returns:
        None (underlying `io_dict` is mutated by the recursion)
    """
    if isinstance(cli_element, (click.Group, click.CommandCollection)):
        element_name = cli_element.name or "kedro"
        io_dict[element_name] = {}
        for command_name in cli_element.list_commands(ctx):
            _recurse_cli(
                cli_element.get_command(ctx, command_name),
                ctx,
                io_dict[element_name],
                get_help,
            )
        if not get_help:
            nested_parameter_list = [
                option.opts for option in cli_element.get_params(ctx)
            ]
            for item in (item for sublist in nested_parameter_list for item in sublist):
                if item not in io_dict[element_name]:
                    io_dict[element_name][item] = None

    elif isinstance(cli_element, click.Command):
        if get_help:  # gets formatted CLI help incl params for printing
            io_dict[cli_element.name] = cli_element.get_help(ctx)
        else:  # gets params for structure purposes
            nested_parameter_list = [
                option.opts for option in cli_element.get_params(ctx)
            ]
            io_dict[cli_element.name] = dict.fromkeys(
                [item for sublist in nested_parameter_list for item in sublist], None
            )


def _get_cli_structure(
    cli_obj: click.Command | (click.Group | click.CommandCollection),
    get_help: bool = False,
) -> dict[str | None, Any]:
    """Code copied over from kedro.tools.cli to maintain backwards compatibility
    with previous versions of kedro (<0.17.5).
    Convenience wrapper function for `_recurse_cli` to work within
    `click.Context` and return a `dict`.
    """
    output: dict[str | None, Any] = {}
    with click.Context(cli_obj) as ctx:  # type: ignore
        _recurse_cli(cli_obj, ctx, output, get_help)
    return output


def _mask_kedro_cli(cli: click.CommandCollection, command_args: list[str]) -> list[str]:
    """Takes a dynamic vocabulary (based on `KedroCLI`) and returns
    a masked CLI input"""
    output = []
    arg_index = 0
    cmd = command_args[0] if command_args else ""
    if cmd in {"--help", "--version", "-h", "-v", ""}:
        return command_args
    click_cmd = cli.get_command(ctx=None, cmd_name=cmd)  # type: ignore
    if click_cmd is None:
        return [MASK]

    current_CLI = _get_cli_structure(click_cmd)
    while (
        arg_index < len(command_args)
        and not command_args[arg_index].startswith("-")
        and command_args[arg_index] in current_CLI
    ):
        output.append(command_args[arg_index])
        current_CLI = current_CLI[command_args[arg_index]]
        arg_index += 1

    # Mask everything except parameter keywords
    for arg in command_args[arg_index:]:
        if arg.startswith("-"):
            if "=" in arg:
                arg_left = arg.split("=")[0]
                if arg_left in current_CLI:
                    output.append(arg_left)
                output.append(MASK)
            elif arg in current_CLI:
                output.append(arg)
            else:
                output.append(MASK)
        else:
            output.append(MASK)

    return output
